# coding=utf-8
import os.path as op
import bloscpack
from bloscpack.defaults import DEFAULT_CHUNK_SIZE
from jagged.npy_backend import JaggedByNPY
from bloscpack.numpy_io import pack_ndarray_file, unpack_ndarray_file


class JaggedByBloscpack(JaggedByNPY):

    def __init__(self,
                 path=None,
                 journal=None,
                 # blosc
                 clevel=5,
                 shuffle=True,
                 cname='lz4hc',
                 # bloscpack
                 chunk_size=DEFAULT_CHUNK_SIZE,
                 offsets=False,
                 checksum='None'):
        super(JaggedByBloscpack, self).__init__(path, journal=journal)
        self.clevel = clevel
        self.shuffle = shuffle
        self.cname = cname
        self.offsets = offsets
        self.checksum = checksum
        self.chunk_size = chunk_size
        self._bp_args = None
        self._blosc_args = None

    def _dest_file(self, index):
        return op.join(self._shards[index % 256], '%d.blp' % index)

    def _read_one(self, key):
        return unpack_ndarray_file(self._dest_file(key))

    def _append_hook(self, data):
        if self._bp_args is None:
            self._bp_args = bloscpack.BloscpackArgs(offsets=self.offsets,
                                                    checksum=self.checksum)
        if self._blosc_args is None:
            self._blosc_args = bloscpack.BloscArgs(typesize=self.dtype.itemsize,
                                                   clevel=self.clevel,
                                                   shuffle=self.shuffle,
                                                   cname=self.cname)
        pack_ndarray_file(data, self._dest_file(self.narrays),
                          chunk_size=self.chunk_size,
                          blosc_args=self._blosc_args,
                          bloscpack_args=self._bp_args)



# coding=utf-8
import joblib
from jagged.pickle_backend import JaggedByPickle


class JaggedByJoblib(JaggedByPickle):

    def __init__(self, path=None, journal=None, arrays_per_chunk=1000, compress=False):
        super(JaggedByJoblib, self).__init__(path, journal, arrays_per_chunk, compress)

    def _load(self, path):
        self._cache = joblib.load(path)

    def _dump(self, path):
        compress = 0 if not self.compress else (5 if self.compress is True else self.compress)
        joblib.dump(self._cache, path, compress=compress)  # cache_size



# coding=utf-8
"""Tests the raw storers."""
from __future__ import print_function, absolute_import, unicode_literals
from future.builtins import zip
import os.path as op
import bcolz
from .fixtures import *
from jagged.base import retrieve_contiguous
from jagged.misc import ensure_dir


# -- lifecycle tests

def test_lifecycle(jagged_raw):
    jagged_raw, path = jagged_raw
    data0 = np.zeros((2, 10))
    data1 = np.ones((3, 10))
    expected = np.vstack((data0, data1))
    with jagged_raw(path=path) as jr:
        # before writing, everything is unknown
        assert jr.shape is None
        assert jr.ndims is None
        assert jr.dtype is None
        # first write-up
        jr.append(data0)
        assert jr.shape == data0.shape
        assert jr.dtype == data0.dtype
        assert jr.ndims == data0.ndim
        assert jr.narrays == 1
        assert len(jr) == len(data0)
        # first read
        assert np.allclose(data0, jr.get()[0])
        # even if we close it...
        jr.close()
        # we can now know shapes and the like
        assert jr.shape == data0.shape
        assert jr.dtype == data0.dtype
        assert jr.ndims == data0.ndim
        # we can reread...
        assert np.allclose(data0, jr.get()[0])
        # we can know shapes and the like
        assert jr.shape == data0.shape
        assert jr.dtype == data0.dtype
        assert jr.ndims == data0.ndim
        # we can append more
        jr.append(data1)
        assert jr.shape == expected.shape
        assert jr.dtype == expected.dtype
        assert jr.ndims == expected.ndim
        assert jr.narrays == 2
        assert len(jr) == len(data0) + len(data1)
        # and the data will be properlly appended
        assert np.allclose(data0, jr.get()[0])
        assert np.allclose(data1, jr.get()[1])


# -- Tests retrieve contiguous

def test_retrieve_contiguous(mock_jagged_raw, columns, contiguity):

    originals, ne, nc, dtype, segments, reader, rng = mock_jagged_raw

    if columns is not None:
        originals = [o[:, tuple(columns)] for o in originals]

    # sanity checks for wrong inputs
    with pytest.raises(ValueError) as excinfo:
        retrieve_contiguous(segments, columns, reader, dtype, ne, nc, 'wrong')
    assert 'Unknown contiguity scheme:' in str(excinfo.value)

    with pytest.raises(ValueError) as excinfo:
        retrieve_contiguous([(-1, 1)], columns, reader, dtype, ne, nc, contiguity)
    assert 'Out of bounds query (base=-1, size=1' in str(excinfo.value)

    with pytest.raises(ValueError) as excinfo:
        retrieve_contiguous([(0, 100000)], columns, reader, dtype, ne, nc, contiguity)
    assert 'Out of bounds query (base=0, size=100000' in str(excinfo.value)

    # insertion order
    views = retrieve_contiguous(segments, columns, reader, dtype, ne, nc, contiguity)
    for o, v in zip(originals, views):
        assert np.allclose(o, v)

    # random order
    o_s = list(zip(originals, segments))
    rng.shuffle(o_s)
    originals, segments = zip(*o_s)
    views = retrieve_contiguous(segments, columns, reader, dtype, ne, nc, contiguity)
    for o, v in zip(originals, views):
        assert np.allclose(o, v)


# -- roundtrip tests

def test_roundtrip(jagged_raw, dataset, columns):
    jagged_raw, path = jagged_raw
    jagged_raw = partial(jagged_raw, path=path)
    rng, originals, ncol = dataset

    # Write
    keys = []
    with jagged_raw() as jr:
        total = 0
        assert jr.dtype is None
        assert jr.shape is None
        for original in originals:
            key = jr.append(original)
            assert jr.is_writing
            keys.append(key)
            total += len(original)
            assert len(jr) == total
        assert jr.dtype == originals[0].dtype
        assert jr.shape == (total, ncol)

    # Read
    def test_read(originals, keys):

        if columns is not None:
            originals = [o[:, columns] for o in originals]

        # test read, one by one
        with jagged_raw() as jr:
            for original, key in zip(originals, keys):
                roundtripped = jr.get([key], columns=columns)[0]
                assert np.allclose(original, roundtripped)

        # test read, in a batch
        with jagged_raw() as jr:
            for original, roundtripped in zip(originals, jr.get(keys, columns=columns)):
                assert np.allclose(original, roundtripped)

    # read all
    # with jagged_raw() as jr:
    #     for original, roundtripped in zip(originals, jr.get(columns=columns)):
    #         original = original if columns is None else original[:, columns]
    #         assert np.allclose(original, roundtripped)

    # read in insertion order
    test_read(originals, keys)

    # read in random order
    or_s = list(zip(originals, keys))
    rng.shuffle(or_s)
    originals, keys = zip(*or_s)
    test_read(originals, keys)


# --- Test self-identification

def test_whatid():
    assert "JaggedByCarray(chunklen=1000," \
           "contiguity=None," \
           "cparams=cparams(clevel=3,cname='zlib',shuffle=False)," \
           "expectedlen=None)" \
           == JaggedByCarray(chunklen=1000,
                             cparams=bcolz.cparams(clevel=3, cname='zlib', shuffle=False),
                             expectedlen=None).what().id()
    assert "JaggedByH5Py(checksum=False," \
           "chunklen=1000," \
           "compression='lzf'," \
           "compression_opts=0," \
           "contiguity=None," \
           "shuffle=True)" \
           == JaggedByH5Py(chunklen=1000,
                           compression='lzf',
                           compression_opts=0,
                           shuffle=True).what().id()


# --- Test factories / curries/ partials

def test_copyconf(jagged_raw):
    jagged_raw, path = jagged_raw
    assert jagged_raw().what().id() == jagged_raw().copyconf()().what().id(), \
        'factory without parameters should give the same config as the constructor'


# --- Misc tests

def test_nonvalid_appends(jagged_raw):
    jagged_raw, path = jagged_raw
    with jagged_raw(path=path) as jr:
        with pytest.raises(Exception) as excinfo:
            jr.append(np.zeros((10, 0)))
        assert 'Cannot append data with sizes 0 in non-leading dimension' in str(excinfo.value)


def test_no_inmemory_storage(jagged_raw):
    # maybe one day we allow these...
    jagged_raw, path = jagged_raw
    with jagged_raw(path=None) as jr:
        with pytest.raises(Exception) as excinfo:
            jr.append(np.zeros((1, 1)))
        assert 'In-memory only arrays are not implemented' in str(excinfo.value)


def test_copy_from(jagged_raw):
    jagged_raw, path = jagged_raw
    path0 = ensure_dir(op.join(path, 'test0'))
    path1 = ensure_dir(op.join(path, 'test1'))
    with jagged_raw(path0) as jr0, jagged_raw(path1) as jr1:
        jr0.append(np.zeros((2, 10)))
        jr0.append(np.ones((3, 10)))
        jr1.append_from(jr0)
        assert np.allclose(jr0.get()[0], jr1.get()[0])


def test_chunked_copy_from(jagged_raw):
    jagged_raw, path = jagged_raw
    path0 = ensure_dir(op.join(path, 'test0'))
    path1 = ensure_dir(op.join(path, 'test1'))
    with jagged_raw(path0) as jr0, jagged_raw(path1) as jr1:
        for _ in range(10):
            jr0.append(np.zeros((2, 10)))
            jr0.append(np.ones((3, 10)))
        jr1.append_from(jr0, arrays_per_chunk=2)
        assert np.allclose(jr0.get()[0], jr1.get()[0])
        with pytest.raises(ValueError) as excinfo:
            jr1.append_from(jr0, arrays_per_chunk=-1)
        assert 'arrays_per_chunk must be None or bigger than 0, it is -1' in str(excinfo.value)


def test_mmap_check_sizes(tmpdir):
    dest = str(tmpdir)
    x = np.empty((5, 2), dtype=np.int32)
    with JaggedByMemMap(dest) as jbm:
        jbm.append(x)
        mmf = jbm._mmpath
    # write row-sized junk
    with open(mmf, 'a') as writer:
        writer.write(str('junk' * 10))
    with JaggedByMemMap(dest) as jbm:
        with pytest.raises(Exception) as excinfo:
            jbm.get([(0, 2)])
        assert 'the number or rows inferred by file size does not coincide' in str(excinfo.value)
    # write junk that look like leftovers of an aborted write
    with open(mmf, 'a') as writer:
        writer.write(str('jagged'))
    with JaggedByMemMap(dest) as jbm:
        with pytest.raises(Exception) as excinfo:
            jbm.get([(0, 2)])
        assert 'the memmap file has incomplete data' in str(excinfo.value)
    # make the memmap way too small
    with open(mmf, 'w') as writer:
        writer.write(str('jagged'))
    with pytest.raises(Exception) as excinfo:
        jbm.get([(0, 2)])
    assert 'mmap length is greater than file size' in str(excinfo.value)



# coding=utf-8
"""Backend using python/numpy mmap bindings."""
from __future__ import absolute_import, unicode_literals, print_function, division
import os.path as op
from future.utils import PY3
import numpy as np
from jagged.base import LinearRawStorage


class JaggedByMemMap(LinearRawStorage):
    """Provides numpy arrays as views of an underlying memmapped array."""

    def __init__(self, path=None, journal=None, contiguity=None, autoviews=True):
        super(JaggedByMemMap, self).__init__(path, journal=journal, contiguity=contiguity)

        if self._path is not None:
            self._mmpath = op.join(self._path, 'data.mm')

        self._mm = None  # numpy memmap for reading / file handler for writing

        self.autoviews = autoviews

    # --- Read

    def _open_read(self):
        if self._mm is None:
            self._mm = np.memmap(self._mmpath,
                                 dtype=self.dtype, shape=self.shape, order=self.order,
                                 mode='r')
        self._check_sizes()

    def _get_hook(self, base, size, columns, dest):
        view = self._mm[base:base+size]
        if columns is not None:
            view = view[:, tuple(columns)]
        if dest is None:
            return view.copy() if not self.autoviews else view
        dest[:] = view
        return dest

    # --- Write

    def _open_write(self, data=None):
        self._mm = open(self._mmpath, mode='a')

    def _append_hook(self, data):
        self._mm.buffer.write(data.data) if PY3 else self._mm.write(str(data.data))

    # --- Lifecycle

    @property
    def is_writing(self):
        return self.is_open and not self.is_reading

    @property
    def is_reading(self):
        return isinstance(self._mm, np.memmap)

    @property
    def is_open(self):
        return self._mm is not None

    def close(self):
        if self.is_writing:
            self._mm.close()
        self._mm = None

    # --- Storage for underlying array shape, dtype, row/column order

    def _len_by_filelen(self):
        """Helps to check sanity of the array."""
        mmsize_bytes = op.getsize(self._mmpath)
        row_size_bytes = self.shape[1] * self.dtype.itemsize
        num_rows = mmsize_bytes // row_size_bytes
        leftovers = mmsize_bytes % row_size_bytes
        return num_rows, leftovers

    def _check_sizes(self):
        if op.isfile(self._mmpath) and self.shape is not None:
            num_rows, leftovers = self._len_by_filelen()
            if 0 != leftovers:
                raise Exception('the memmap file has incomplete data '
                                '(%d leftover bytes from a partially written array).'
                                '(are you missing transactions?)' % leftovers)
            if num_rows != self.shape[0]:
                raise Exception('the number or rows inferred by file size '
                                'does not coincide with the length of the store '
                                '(%d != %d)' % (num_rows, self.shape[0]))

#
# Remember that resize for numpy mmap objects never resize the file under the hood.
# Numpy mmap ndarray subclass code is simple and neat, can be read in no time. Get back there.
#
# Document that when using mode 'auto', everything returned is a view to the large memmapped array.
#



# coding=utf-8

import bcolz
from whatami import whatable
import os.path as op
from .base import LinearRawStorage


class JaggedByCarray(LinearRawStorage):
    """
    A Jagged store that uses in-disk `bcolz.carray` to store the data.

    This backend should be good for compressable data accessed sequentially or in batch range queries.
    Random access of small segments will suffer from a considerable performance degradation.

    Usually these stores are backed by many files, so access via network file systems or from spin disks can
    potentially be inefficient.

    Parameters
    ----------
    path : string
      the carray will/must reside here

    contiguity : string, default None
      see base class

    journal : must quack like JaggedJournal, default None
      see base class

    expectedlen : int, default None
      passed to the carray on creation, the expected number of rows in the store
      carray will use it to guess a good chunksize
      the actual size of each chunk will of course depend also on the number of columns
      must be None if `chunklen` is provided

    chunklen : int, default None
      passed to the carray on creation, the number of rows to store per chunk
      the actual size of each chunk will of course depend also on the number of columns
      must be None if `expectedlen` is provided

    cparams : `bcolz.cparams`, default bcolz.cparams(clevel=5, shuffle=False, cname='lz4hc')
      the compression configuration for bcolz; only used if the array is empty
    """

    def __init__(self,
                 path=None,
                 journal=None,
                 contiguity=None,
                 # bcolz params
                 expectedlen=None,
                 chunklen=1024 ** 2 // 2,  # 500K rows
                 cparams=bcolz.cparams(clevel=5, shuffle=False, cname='lz4hc')):

        super(JaggedByCarray, self).__init__(path, journal=journal, contiguity=contiguity)

        self.expectedlen = expectedlen
        self.chunklen = chunklen
        self.cparams = whatable(cparams, add_properties=True)
        self._bcolz = None

    def _bcolz_dir(self):
        # Needs to be different than self._path or metainfo gets deleted
        return op.join(self.path_or_fail(), 'bcolz')

    # --- Write

    def _open_write(self, data=None):
        if self._bcolz is None:
            try:  # append
                self._bcolz = \
                    bcolz.carray(None,
                                 rootdir=self._bcolz_dir(),
                                 mode='a',
                                 # bcolz conf in case mode='a' semantics change to create, otherwise innocuous
                                 chunklen=self.chunklen,
                                 expectedlen=self.expectedlen,
                                 cparams=self.cparams)
            except:  # create
                self._bcolz = \
                    bcolz.carray(data[0:0],
                                 rootdir=self._bcolz_dir(),
                                 mode='w',
                                 chunklen=self.chunklen,
                                 expectedlen=self.expectedlen,
                                 cparams=self.cparams)

    def _append_hook(self, data):
        self._bcolz.append(data)

    # --- Read

    def _open_read(self):
        if self._bcolz is None:
            self._bcolz = bcolz.carray(None, rootdir=self._bcolz_dir(), mode='r')

    def _get_hook(self, base, size, columns, dest):
        if dest is not None and columns is None:
            # measure if this has any performance benefit, if so, asks for it to be public API
            self._bcolz._getrange(base, size, dest)
            return dest
        if columns is not None:
            view = self._bcolz[base:base+size, columns]
        else:
            view = self._bcolz[base:base+size]
        if dest is not None:
            dest[:] = view
            return dest
        return view

    # --- Lifecycle

    @property
    def is_writing(self):
        return self.is_open and self._bcolz.mode in ('w', 'a')

    @property
    def is_reading(self):
        return self.is_open and self._bcolz.mode == 'r'

    @property
    def is_open(self):
        return self._bcolz is not None

    def close(self):
        if self.is_writing:
            self._bcolz.flush()
        self._bcolz = None



# coding=utf-8
import os.path as op

import numpy as np
import h5py

from jagged.base import LinearRawStorage


class JaggedByH5Py(LinearRawStorage):

    def __init__(self,
                 path=None,
                 journal=None,
                 contiguity=None,
                 # hdf params
                 dset_name='data',
                 chunklen=None,
                 compression=None,
                 compression_opts=None,
                 shuffle=False,
                 checksum=False):
        super(JaggedByH5Py, self).__init__(path, journal=journal, contiguity=contiguity)

        self._dset_name = dset_name

        if self._path is not None:
            self._h5_path = op.join(self._path, 'data.h5')
        self._h5 = None
        self._dset = None

        self.chunklen = chunklen
        self.compression = compression
        self.compression_opts = compression_opts
        self.shuffle = shuffle
        self.checksum = checksum

    # --- Read

    def _open_read(self):
        if self._h5 is None:
            self._h5 = h5py.File(self._h5_path, mode='r')
            self._dset = self._h5[self._dset_name]

    def _get_hook(self, base, size, columns, dest):

        # h5py does not handle graciously this case
        if size == 0:
            if dest is not None:
                return dest
            nc = len(columns) if columns is not None else self._dset.shape[-1]
            return np.empty((0, nc), dtype=self._dset.dtype)

        # easy case, no column subset requested
        if columns is None:
            if dest is None:
                return self._dset[base:base+size]  # should we force read with [:]? add to benchmark
            else:
                self._dset.read_direct(dest, source_sel=np.s_[base:base+size])
                return dest

        # N.B.: tuple(columns) to force 2d if columns happens to be a one-element list
        # column-subset is requested
        # h5py only supports increasing order indices in fancy indexing
        #   https://github.com/h5py/h5py/issues/368
        #   https://github.com/h5py/h5py/issues/368
        # (boiling down to issues with hdf5 hyperslabs)

        if not np.any(np.diff(columns) < 1):
            if dest is not None:
                self._dset.read_direct(dest, source_sel=np.s_[base:base+size, tuple(columns)])
                return dest
            else:
                return self._dset[base:base+size, tuple(columns)]

        # better slow than unsupported...
        columns, inverse = np.unique(columns, return_inverse=True)
        if dest is not None:
            dest[:] = self._dset[base:base+size, tuple(columns)][:, inverse]
            return dest
        else:
            return self._dset[base:base+size, tuple(columns)][:, inverse]

    # --- Write

    def _open_write(self, data=None):
        if self._h5 is None:
            self._h5 = h5py.File(self._h5_path, mode='a')
            if 'data' not in self._h5:
                # http://docs.h5py.org/en/latest/high/dataset.html
                chunks = None
                if self.chunklen is not None:
                    chunks = (self.chunklen,) + (data.shape[1:] if data.ndim > 1 else ())
                self._dset = self._h5.create_dataset(self._dset_name,
                                                     dtype=data.dtype,
                                                     shape=(0, data.shape[1]),
                                                     maxshape=(None, data.shape[1]),
                                                     chunks=chunks,
                                                     compression=self.compression,
                                                     compression_opts=self.compression_opts,
                                                     shuffle=self.shuffle,
                                                     fletcher32=self.checksum)
            else:
                self._dset = self._h5[self._dset_name]

    def _append_hook(self, data):
        base = len(self)
        size = len(data)
        self._dset.resize(base + size, axis=0)
        self._dset[base:(base+size)] = data

    # --- Lifecycle

    @property
    def is_writing(self):
        return self.is_open and self._h5.mode != 'r'

    @property
    def is_reading(self):
        return self.is_open and self._h5.mode == 'r'

    @property
    def is_open(self):
        return self._h5 is not None

    def close(self):
        if self._h5 is not None:
            self._h5.close()
            self._h5 = None



# coding=utf-8
"""A jumble of seemingly useful stuff."""
from __future__ import unicode_literals
from itertools import chain
from contextlib import contextmanager
import numbers
import os
import os.path as op
import numpy as np


def home():  # pragma: no cover
    """Returns current user home dir."""
    return op.expanduser('~')  # Valid in both py2 and py3


def ensure_writable_dir(path):  # pragma: no cover
    """Ensures that a path is a writable directory."""
    def check_path(path):
        if not op.isdir(path):
            raise Exception('%s exists but it is not a directory' % path)
        if not os.access(path, os.W_OK):
            raise Exception('%s is a directory but it is not writable' % path)
    if op.exists(path):
        check_path(path)
    else:
        try:
            os.makedirs(path)
        except Exception:
            if op.exists(path):  # Simpler than using a file lock to work on multithreading...
                check_path(path)
            else:
                raise
    return path


def ensure_dir(path):  # pragma: no cover
    return ensure_writable_dir(path)


@contextmanager
def cd(newdir):
    prevdir = os.getcwd()
    os.chdir(op.expanduser(newdir))
    try:
        yield
    finally:
        os.chdir(prevdir)


# --- Intervals

def crossings(x, threshold=0, after=False):
    """Returns the indices of the elements before or after crossing a threshold.

    N.B. touching the threshold itself is considered a cross.

    Parameters
    ----------
    x: array
    The data

    threshold: float, default 0
    Where crossing happens.

    after: bool, default False
    If True, the indices represent the elements after the cross, if False the elements before the cross.

    Returns
    -------
    The indices where crosses happen.

    Examples
    --------

    >>> print(crossings(np.array([0, 1, -1, -1, 1, -1])))
    [0 1 3 4]
    >>> print(crossings(np.array([0, 1, -1, -1, 1, -1]), after=True))
    [1 2 4 5]
    >>> print(crossings(np.array([0, 0, 0])))
    []
    >>> print(crossings(np.array([0, 3, -3, -3, 1]), threshold=1))
    [0 1 3]
    >>> print(crossings(np.array([0, 3, -3, -3]), threshold=-2.5))
    [1]
    >>> print(crossings(np.array([[0, 3], [-3, -3]]), threshold=-2.5))  # doctest: +IGNORE_EXCEPTION_DETAIL
    Traceback (most recent call last):
    Exception: Only 1D arrays, please (you gave me 2 dimensions)
    """
    if len(x.shape) > 1:
        raise Exception('Only 1D arrays, please (you gave me %d dimensions)' % len(x.shape))
    where_crosses = np.where(np.diff(np.sign(x - threshold)))[0]
    if after:
        return where_crosses + 1
    return where_crosses


def find_intervals(x):
    """
    Finds the intervals in which x is True or non-zero.


    Returns
    -------
    Pairs of indices representing the intervals in which x is True or nonzero.
    The pairs represent valid python intervals, lower point included, upper point excluded.


    Examples
    --------
    >>> find_intervals([])
    []
    >>> find_intervals([1])
    [(0, 1)]
    >>> find_intervals([0, 1])
    [(1, 2)]
    >>> find_intervals([0, 0, 1, 1, 0, 0, 1, 1, 0])
    [(2, 4), (6, 8)]
    >>> find_intervals([0, 0, 0])
    []
    >>> find_intervals([1, 1, 1])
    [(0, 3)]
    >>> find_intervals([True, True, True])
    [(0, 3)]
    >>> find_intervals([1, 1, 1, 0])
    [(0, 3)]
    """
    # This ugly 6 lines are here because:
    #   - we allow to pass lists but we need numpy arrays
    #   - we want to allow both boolean (True, False) arrays and numeric arrays
    #   - we want to use the crossings function which only accepts numeric arrays
    if not isinstance(x, np.ndarray):
        x = np.array(x)
    if not x.dtype == np.bool:
        x = x != 0
    zeros_ones = np.zeros_like(x, dtype=np.int)
    zeros_ones[x] = 1

    # Find where we change from being in an interval to not being in an interval
    starts_ends = list(crossings(zeros_ones, after=True))

    # Do we start already in an interval?
    if len(zeros_ones) > 0 and 1 == zeros_ones[0]:
        starts_ends = [0] + starts_ends

    # Do we end in an interval?
    if len(zeros_ones) > 0 and 1 == zeros_ones[-1]:
        starts_ends = starts_ends + [len(x)]

    assert len(starts_ends) % 2 == 0

    starts = starts_ends[0::2]
    ends = starts_ends[1::2]
    return list(zip(starts, ends))


def is_valid_segment(ss, relative_to=None):
    if not isinstance(ss, (tuple, list)):
        return False
    if not len(ss) == 2:
        return False
    ss_base, ss_size = ss
    if not isinstance(ss_base, numbers.Integral) and isinstance(ss_size, numbers.Integral):
        return False
    if relative_to is not None:
        base, size = relative_to
        if ss_base < 0 or (base + ss_base + ss_size) > (base + size):
            return False
    return True


def bool2segments(ss, size):
    ssa = np.array(ss)
    if ssa.dtype.kind == 'b' and ssa.ndim == 1 and len(ssa) == size:
        return [(start, end - start) for start, end in find_intervals(ssa)]
    return None


def subsegments(segment, *subs):
    """Make subsegments relative to the start of a base segment, checking for boundaries.

    Parameters
    ----------
    segment : tuple (base, size)
      The segment to which relative subsegments are being specified

    subs : list of (base, size) boolean arrays specifying subsegments
      These can be either something like (3, 8) (ss_base, ss_size), or boolean lists/arrays
      It is assumed that ss_base is here is offset from `segment` base

    Returns
    -------
    A list of subsegments [(base, size)], each lying within the boundaries of `segment`.

    Examples
    --------
    >>> subsegments((5, 100))
    []
    >>> subsegments((5, 100), (11, 14))
    [(16, 14)]
    >>> subsegments((5, 100), (11, 14), (3, 88))
    [(16, 14), (8, 88)]
    >>> subsegments((0, 5), [True, True, False, True, True])
    [(0, 2), (3, 2)]
    >>> subsegments((0, 5), [False] * 5)
    []
    >>> subsegments((0, 5), [True] * 5)
    [(0, 5)]
    >>> subsegments((0, 5), np.array([True] * 5))
    [(0, 5)]
    >>> subsegments((0, 5), [True, True, False, True, True], (2, 2))
    [(0, 2), (3, 2), (2, 2)]
    >>> subsegments((0, 100), (90, 11))  # doctest: +IGNORE_EXCEPTION_DETAIL
    Traceback (most recent call last):
    ValueError: (90, 11) is not a valid subsegment specification for (0, 100)
    >>> subsegments((0, 100), (-3, 8))  # doctest: +IGNORE_EXCEPTION_DETAIL
    Traceback (most recent call last):
    ValueError: (90, 11) is not a valid subsegment specification for (0, 100)
    >>> subsegments((0, 100), ('a', 8))  # doctest: +IGNORE_EXCEPTION_DETAIL
    Traceback (most recent call last):
    ValueError: ('a', 8) is not a valid subsegment specification for (0, 100)
    >>> subsegments((0, 100), 'crazyyou')  # doctest: +IGNORE_EXCEPTION_DETAIL
    Traceback (most recent call last):
    ValueError: 'crazyyou' is not a valid subsegment specification for (0, 100)
    """

    # This implementation is slow, but seemingly correct; I do not think it will bottleneck

    base, size = segment

    def bool_and_valid(ss):
        if is_valid_segment(ss, relative_to=segment):
            return [ss]
        ss_from_bool = bool2segments(ss, size)
        if ss_from_bool is not None and all(is_valid_segment(ss, relative_to=segment) for ss in ss_from_bool):
            return ss_from_bool
        raise ValueError('%r is not a valid subsegment specification for %r' % (ss, segment))

    return [(base + ss_base, ss_size) for ss_base, ss_size in
            chain.from_iterable(bool_and_valid(ss) for ss in subs)]



# coding=utf-8
import gzip
from operator import itemgetter
from jagged.base import JaggedRawStore
import os.path as op
try:
    import cPickle as pickle
except ImportError:  # pragma: no cover
    import pickle


class JaggedByPickle(JaggedRawStore):
    """A chunked store based on pickle."""

    def __init__(self, path=None, journal=None, arrays_per_chunk=1000, compress=False):
        super(JaggedByPickle, self).__init__(path, journal)
        self.arrays_per_chunk = arrays_per_chunk
        self.compress = compress
        self._cache = []
        self._cached_pickle_num = None
        self._writing = None

    # --- Pickles

    def _dump(self, path):
        with gzip.open(path, 'wb') if self.compress else open(path, 'wb') as writer:
            pickle.dump(self._cache, writer, protocol=2)
            # protocol=2 instead of highest to maintain py2 compat of the store

    def _load(self, path):
        with gzip.open(path, 'rb') if self.compress else open(path, 'rb') as reader:
            self._cache = pickle.load(reader)

    def _pickle_num(self, index):
        return index // self.arrays_per_chunk

    def _pickle_file(self, index):
        path = op.join(self.path_or_fail(), '%d.pkl' % self._pickle_num(index))
        return (path + '.gz') if self.compress else path

    def _save_pickle(self):
        if self.is_writing:
            self._dump(self._pickle_file(self.narrays))

    def _read_pickle(self, index):
        pickle_num = self._pickle_num(index)
        if self._cached_pickle_num != pickle_num:
            try:
                self._load(self._pickle_file(index))
            except IOError:
                self._cache = []
            self._cached_pickle_num = pickle_num

    # --- Cache

    def _cache_full(self):
        return self._cache is not None and len(self._cache) == self.arrays_per_chunk

    # --- Read

    def _open_read(self):
        self._writing = False

    def _get_views(self, keys, columns):
        if keys is None:
            keys = range(self.narrays)

        keys = [(key, order) for order, key in enumerate(keys)]

        views = []
        for key, order in sorted(keys):
            if not 0 <= key < self.narrays:
                raise ValueError('Key not in storage: %d' % key)
            self._read_pickle(key)
            array = self._cache[key % self.arrays_per_chunk]
            if columns is not None:
                array = array[:, tuple(columns)]
            views.append((array, order))
        views = list(map(itemgetter(0), sorted(views, key=itemgetter(1))))

        return views

    # --- Write

    def _open_write(self, data=None):
        self._writing = True
        self._read_pickle(self.narrays)

    def _append_hook(self, data):
        self._cache.append(data.copy())
        if self._cache_full():
            self._save_pickle()
            self._cache = []
            self._cached_pickle_num += 1

    # --- Lifecycle

    @property
    def is_open(self):
        return self._writing is not None

    @property
    def is_writing(self):
        return self.is_open and self._writing

    @property
    def is_reading(self):
        return self.is_open and not self._writing

    def close(self):
        self._save_pickle()
        self._cache = None
        self._cached_pickle_num = None
        self._writing = None

#
# We let pandas decide how to pickle
# In general it would use protocol 2, so pickles can be read also in py2
# We can be as clever as we want with caches: many read caches with LRU, one write cache...
#



from mmap import mmap, ACCESS_READ
from operator import itemgetter
import os.path as op

from future.builtins import range

from jagged.base import JaggedRawStore, JaggedJournal
from jagged.compression.compressors import BloscCompressor
from whatami import What


class JaggedByBlosc(JaggedRawStore):

    # Memmapped
    # Not chunked - hope to keep using bcolz for that

    def __init__(self, path=None, journal=None, compressor=BloscCompressor):
        super(JaggedByBlosc, self).__init__(path, journal=journal)
        self.compressor = compressor
        self._mm = None
        self._writing = None
        self._bjournal = None

    def _bytes_journal(self):
        if self._bjournal is None:
            self._bjournal = JaggedJournal(op.join(self.path_or_fail(), 'meta', 'bytes_journal'))
        return self._bjournal

    def _compressor(self):
        if not isinstance(self.compressor, BloscCompressor):
            self.compressor = self.compressor(dtype=self.dtype,
                                              shape=self.shape,
                                              order=self.order)
        return self.compressor

    # --- Custom what() to be available at any circumstance

    def what(self):
        try:
            return What(self.__class__.__name__, {'compressor': self.compressor()})
        except TypeError:
            return What(self.__class__.__name__, {'compressor': self.compressor})

    # --- Write

    def _open_write(self, data=None):
        self._mm = open(op.join(self.path_or_fail(), 'data'), 'ab')
        self._writing = True

    def _append_hook(self, data):
        compressor = self._compressor()
        compressed = compressor.compress(data)
        self._mm.write(compressed)
        self._bytes_journal().append(compressed)

    # --- Read

    def _open_read(self):
        self._mm = open(op.join(self.path_or_fail(), 'data'), 'r')
        self._mm = mmap(self._mm.fileno(), 0, access=ACCESS_READ)
        self._writing = False

    def _read_segment(self, base, size):
        return self._mm[base:base+size]

    def _get_views(self, keys, columns):

        if keys is None:
            keys = range(self.narrays)

        keys = [(key, order) for order, key in enumerate(keys)]

        compressor = self._compressor()
        views = []
        for key, order in sorted(keys):
            base, size = self._bytes_journal().base_size(key)  # cache these segments?
            array = compressor.decompress(self._read_segment(base, size))
            if columns is not None:
                array = array[:, tuple(columns)]
            views.append((array, order))
        views = list(map(itemgetter(0), sorted(views, key=itemgetter(1))))

        return views

    # --- Lifecycle

    @property
    def is_reading(self):
        return self.is_open and not self.is_writing

    @property
    def is_writing(self):
        return self.is_open and self._writing

    @property
    def is_open(self):
        return self._mm is not None

    def close(self):
        if self.is_open:
            self._mm.close()
        self._mm = None
        self._writing = None



# coding=utf-8
from future.builtins import range
from itertools import chain
import os.path as op
import os
import numpy as np
from jagged.base import JaggedRawStore
from jagged.misc import ensure_dir


class JaggedByNPY(JaggedRawStore):
    """Stores each array in an individual .npy file."""

    def __init__(self, path=None, journal=None):
        super(JaggedByNPY, self).__init__(path, journal=journal)
        self._shards = None
        if path is not None:
            self._all_shards()

    # We can do this memory map (see np.load)

    def _all_shards(self):
        if self._shards is None:
            self._shards = [ensure_dir(op.join(self.path_or_fail(), str(shard))) for shard in range(256)]
        return self._shards
        # random note, 256 is the last cached int in cpython

    def _dest_file(self, index):
        return op.join(self._shards[index % 256], '%d.npy' % index)

    def _infer_numarrays(self):
        numarrays = 0
        for shard in self._shards:
            numarrays = max(chain([numarrays], (int(fn[:-4]) + 1 for fn in os.listdir(shard))))
        return numarrays

    def check_numarrays(self):
        assert self._infer_numarrays() == self.journal().numarrays()

    # --- Write

    def _open_write(self, data=None):
        pass

    def _append_hook(self, data):
        np.save(self._dest_file(self.narrays), data)

    # --- Read

    def _open_read(self):
        pass

    def _read_one(self, key):
        return np.load(self._dest_file(key))

    def _get_one(self, key, columns):
        data = self._read_one(key)
        if columns is not None:
            data = data[:, tuple(columns)]
        return data

    def _get_views(self, keys, columns):
        if keys is None:
            return list(self._get_one(key, columns) for key in range(self.journal().numarrays()))
        return [self._get_one(key, columns) for key in keys]

    # --- Lifecycle

    @property
    def is_writing(self):
        return True

    @property
    def is_reading(self):
        return True

    @property
    def is_open(self):
        return True

    def close(self):
        pass



# coding=utf-8
"""Benchmarking utilities.
Some of these are inspired by bloscpack / bloscpack-benchmarks.
  https://github.com/Blosc/bloscpack-benchmarking
"""
import os.path as op
import socket
import datetime
import subprocess
import json
import os
from jagged.misc import ensure_dir
import psutil

#
# Timing is hard and we should at least use timeit
# (something with support for calibration and repetition).
# A great resource is also pytest benchmark
#   https://pypi.python.org/pypi/pytest-benchmark/2.5.0
#   https://bitbucket.org/haypo/misc/src/tip/python/benchmark.py
# There are a bunch of benchmarker / timer etc. libraries in pypi
# Do not forget about /usr/bin/time -v
#


def timestr():
    return datetime.datetime.now().strftime('%Y-%m-%d-%H-%M-%S')


def hostname():
    return socket.gethostname()


def collect_sysinfo(dest=None):
    """
    Collects basic information from the machine using several tools.
    This needs to run as root.
    Note that speeds are theoretical, not measured
    (specially peak network and network drives speeds should be measured).

    Prerequisites
    -------------
    If in ubuntu:
      sudo apt-get install smartmontools inxi dmidecode
    If in arch:
     sudo pacman -S smartmontools inxi dmidecode

    What is run
    -----------
    # Basic information about mount points
    mount > mount.info
    # Inxi reports
    inxi > inxi.info
    # Full dmidecode
    dmidecode > dmidecode.info
    # Network speed information
    dmesg | grep -i duplex > network-speed.info
    # SMART information
    sudo smartctl -a /dev/sda > smartctl-sda.info

    References
    ----------
    http://www.binarytides.com/linux-commands-hardware-info/
    http://www.cyberciti.biz/faq/linux-command-to-find-sata-harddisk-link-speed/
    http://www.cyberciti.biz/faq/howto-setup-linux-lan-card-find-out-full-duplex-half-speed-or-mode/
    http://www.cyberciti.biz/tips/linux-find-out-wireless-network-speed-signal-strength.html
    """

    #
    # Any way of getting actual memory latencies, CAS...?
    # Also we could look at pure python libraries like dmidecode
    #

    if dest is None:
        dest = op.join(op.dirname(__file__), 'sysinfo')
    dest = op.join(ensure_dir(op.join(dest, hostname())), timestr() + '.json')

    info = {
        'mount': subprocess.check_output('mount'),
        'dmesg-eth': '\n'.join(line for line in subprocess.check_output('dmesg').splitlines() if 'duplex' in line),
        'iwconfig': subprocess.check_output('iwconfig'),
        'inxiF': subprocess.check_output(['inxi', '-c 0', '-F']),
        # add some more inxi stuff
    }

    with open(dest, 'w') as writer:
        json.dump(info, writer, indent=2, sort_keys=True)


def du(path):
    """Returns the size of the tree under path in bytes."""
    return int(subprocess.check_output(['du', '-s', '-L', '-B1', path]).split()[0].decode('utf-8'))


def drop_caches(path, drop_level=3, max_size='1000G', verbose=False):
    #
    # Some light reading
    #   http://www.linuxatemyram.com/play.html
    # vmtouch
    #   https://aur.archlinux.org/packages/vmtouch/
    #   http://serverfault.com/questions/278454/is-it-possible-to-list-the-files-that-are-cached
    #   http://serverfault.com/questions/43383/caching-preloading-files-on-linux-into-ram
    # fincore
    #   yaourt -S --noconfirm perl-file-sharedir-install
    #   yaourt -S --noconfirm fincore
    # To drop system caches, one needs root;
    # an option, add the program to sudoers so no pass is required.
    #
    if 0 != os.system('vmtouch -e -f -q -m %s "%s"' % (max_size, path)):
        if os.geteuid() == 0:
            os.system('echo %d > /proc/sys/vm/drop_caches' % drop_level)
            if verbose:
                print('Full system cache dropped because of %s' % path)
        else:
            raise RuntimeError('Need vmtouch or root permission to drop caches')
    else:
        if verbose:
            print('All pages under %s evicted' % path)


def sync():
    """Flushes buffers to disk."""
    os.system('sync')


def available_ram():
    return psutil.virtual_memory().available

#
# We need to make sure that:
#  - we go beyond microbenchmarks and look at relevant tasks
#    e.g. realtime visualisation or data exploration as opposed to batch
#

#
# Measure dataset complexity (e.g. lempel ziv via compression) and report it
#



# coding=utf-8
"""Convenient (and somehow performing) storage of objects with homogeneous types but different lengths.

"jagged" array providers have very simple, low level contracts:
 - Focus on reading performance, append only store.
 - Use numpy arrays as the canonical data carriers
 - May or may not restrict the type of the stored elements
 - Retrieve only by providing indices *collections*
   No explicit support for slice notation
 - All clases are whatami whatables
"""
from __future__ import absolute_import, unicode_literals, print_function
from future.utils import bytes_to_native_str
from abc import ABCMeta
from array import array
from functools import partial
import os.path as op
from operator import itemgetter
import json

from future.builtins import range, map
from toolz import merge, partition_all
import numpy as np

from jagged.misc import ensure_dir
from whatami import whatable

try:
    import cPickle as pickle
except ImportError:  # pragma: no cover
    import pickle


# --- Journals (persitence of array lengths)


def _int_or_0(v):
    if v is None:
        return 0
    return int(v)


def _read_full_file(x, path):
    """Reads the full contentes of file path into array x."""
    with open(path, 'rb') as reader:
        reader.seek(0, 2)
        size = reader.tell()
        reader.seek(0, 0)
        if size % x.itemsize != 0:
            raise Exception('Truncated file')
        x.fromfile(reader, size // x.itemsize)
        return x


class JaggedJournal(object):
    """Keeps track and persists information about the sizes of added arrays."""

    # a journal must be instantiated only when jagged knows its location
    # a journal can be shared by many jagged instances (e.g. when storing different columns by different jaggeds)

    def __init__(self, path):
        super(JaggedJournal, self).__init__()
        self._path = ensure_dir(path)
        # base and length of each added array
        self._lengths_file = op.join(self._path, 'lengths.array')
        self._lengths = self._read_lengths()
        self._bases = None
        # total number of rows and arrays
        self._sizes_file = op.join(self._path, 'size.json')
        self._numrows, self._numarrays = self._read_sizes()

    def append(self, data):
        """Appends the data array to the journal."""
        self._add_length(data)
        self._add_sizes(data)

    # --- Num rows, num arrays (redundant with lengths, light and good for redundancy)

    def _add_sizes(self, data):
        """Adds to numrows and numarrays the sizes of the array data and immediatly persists them."""
        self._numrows += len(data)
        self._numarrays += 1
        with open(self._sizes_file, 'w') as writer:
            json.dump({'numrows': self._numrows, 'numarrays': self._numarrays}, writer, indent=2)

    def _read_sizes(self):
        """Reads the current numrows and numarrays values from persistent storage.
        If there is no info stored, makes them 0.
        """
        if op.isfile(self._sizes_file):
            with open(self._sizes_file, 'r') as reader:
                sizes = json.load(reader)
                return _int_or_0(sizes['numrows']), _int_or_0(sizes['numarrays'])
        return 0, 0

    def numrows(self):
        """Returns the total number of rows in the jagged instance."""
        return self._numrows

    def numarrays(self):
        """Returns the number of arrays in the jagged instance."""
        return self._numarrays

    # --- Base and size of each array

    def _add_length(self, data):
        """Adds the length to the journal and immediatly persists it."""
        self._lengths.append(len(data))
        with open(self._lengths_file, 'ab') as writer:
            self._lengths[-1:].tofile(writer)

    def _read_lengths(self):
        """Reads the lengths from persistent storage, if it does not exist, returns an empty array."""
        lengths = array(bytes_to_native_str(b'l'))
        if op.isfile(self._lengths_file):
            _read_full_file(lengths, self._lengths_file)
        return lengths

    def lengths(self):
        """Returns an array with the length of each array added to the journal."""
        return self._read_lengths()

    def bases(self):
        """Returns where each array would start if the storage is linear."""
        if self._bases is None or len(self._bases) < len(self._lengths):
            self._bases = np.hstack(([0], np.cumsum(self._lengths)))
        return self._bases

    def start_end(self, index):
        """Returns the start and end of the array at index."""
        base, size = self.base_size(index)
        return base, base + size

    def base_size(self, index):
        """Returns the base and size of the array at index."""
        return self.bases()[index], self.lengths()[index]

    # --- Sanity checks

    def check_consistency(self):
        """Checks the internal consistency of the journal."""
        assert len(self.lengths()) == len(self.bases())
        assert len(self.lengths()) == self.numarrays()
        assert len(np.sum(self.lengths())) == self.numrows()

# --- Raw stores


@whatable(add_properties=False)
class JaggedRawStore(object):
    """Persistent storage of objects of the same type but different length."""

    __metaclass__ = ABCMeta

    def __init__(self, path, journal=None):
        super(JaggedRawStore, self).__init__()
        self._path = path
        if self._path is not None:
            ensure_dir(self._path)
        self._template = None    # how the saved arrays look like
        self._journal = journal  # sizes of the added arrays

    # --- Where this storage resides

    def path_or_fail(self):
        """Returns the path if set, otherwise raises an exception."""
        if self._path is None:
            raise Exception('In-memory only arrays are not implemented for %s.' % self.what().id())
        return self._path

    # --- Journal

    def journal(self):
        if self._journal is None:
            self._journal = JaggedJournal(op.join(self.path_or_fail(), 'meta', 'journal'))
        return self._journal

    # --- Template

    def template(self):
        template_dir = ensure_dir(op.join(self.path_or_fail(), 'meta', 'template'))
        template_path = op.join(template_dir, 'template.npy')
        if self._template is None:
            if op.isfile(template_path):
                self._template = np.load(template_path)
        return self._template

    def _write_template(self, data):
        template_dir = ensure_dir(op.join(self.path_or_fail(), 'meta', 'template'))
        template_path = op.join(template_dir, 'template.npy')
        np.save(template_path, data[:0])

    def can_add(self, data):
        """Returns True iff data can be stored.
        This usually means it is of the same kind as previously stored arrays.
        """
        # Obviously we could just store arbitrary arrays in some implementations (e.g. NPY)
        # But lets keep jagged contracts...
        template = self.template()
        if template is None:
            return True
        return (template.dtype >= data.dtype and
                data.shape[-1] == template.shape[-1] and
                np.isfortran(data) == np.isfortran(data))

    # --- Lifecycle

    # N.B. at the moment, to make things simple, we only want write or read
    # We should ensure concurrency does not break this rule (only one writer xor many readers)
    # Of course we could use stuff like SWMR from hdf5 or role our own, more featureful and compled concurrency
    # Not a priority

    @property
    def is_writing(self):
        """Returns whether we can append more data using this jagged instance."""
        raise NotImplementedError()

    @property
    def is_reading(self):
        """Returns whether we can append more data using this jagged instance."""
        raise NotImplementedError()

    @property
    def is_open(self):
        """Returns whether we are currently open in any mode."""
        raise NotImplementedError()

    def close(self):
        """Flushes buffers to permanent storage and closes the underlying backend."""
        raise NotImplementedError()

    # --- Writing data

    def _open_write(self, data=None):
        """Opens in writing mode, returns None.

        Parameters
        ----------
        data : numpy array like, default None
          data schema to use by the storage, needed if this is the first opening of the repository
        """
        raise NotImplementedError()

    def append(self, data):
        """Appends new data to this storage.

        If the storage is empty, this will define the dtype of the store.

        Parameters
        ----------
        data : numpy-array like
          The data to append, must have a compatible dtype with what was already added to the store.

        Returns
        -------
        An integer addressing the added array in the storage
        """

        # at the moment we do not allow coordinate-less stores
        self.path_or_fail()

        # check data validity
        if any(s < 1 for s in data.shape[1:]):
            raise Exception('Cannot append data with sizes 0 in non-leading dimension (%s, %r)' %
                            (self.what().id(), data.shape))

        # check we can write
        if self.is_reading and not self.is_writing:
            self.close()

        # template
        if self.template() is None:
            self._write_template(data)
        assert self.can_add(data)

        # id log
        if not op.isfile(op.join(self.path_or_fail(),  'meta', 'whatid.txt')):
            ensure_dir(op.join(self.path_or_fail(),  'meta'))
            with open(op.join(self.path_or_fail(), 'meta', 'whatid.txt'), 'w') as writer:
                writer.write(self.what().id())

        # open
        self._open_write(data)

        # write
        self._append_hook(data)

        # bookkeeping
        index = self.journal().numarrays()
        self.journal().append(data)

        # done
        return index

    def _append_hook(self, data):
        """Saves the data, returns nothing."""
        raise NotImplementedError()

    def append_from(self, jagged, arrays_per_chunk=None):
        """Appends all the contents of jagged."""
        for chunk in jagged.iter_arrays(arrays_per_chunk=arrays_per_chunk):
            for data in chunk:
                self.append(data)

    # --- Reading

    def _open_read(self):
        """Opens in reading mode, returns None."""
        raise NotImplementedError()

    def _get_views(self, keys, columns):
        """Returns a list of arrays corresponding to the provided keys and columns."""
        raise NotImplementedError()

    def get(self, keys=None, columns=None, factory=None):
        """Returns a list with the data specified in `keys` (and `columns`), possibly transformed by `factory`.

        Concrete implementations may warrant things like "all segments actually lie in congiguous regions in memory".

        Parameters
        ----------
        keys : list of keys
          specifies which elements to retrieve; if None, all arrays are returned

        columns : list of integers, default None
          specifies which columns to retrieve; if None, retrieve all columns

        factory : factory(ndarray)->desired type, default None
          transforms each of the returned elements into a desired type (for example, a pandas DataFrame)
          another use can be to apply summary statistics

        Returns
        -------
        A list with the retrieved elements, possibly transformed by factory.
        """

        # at the moment we do not allow coordinate-less stores
        self.path_or_fail()

        # flush if needed
        if self.is_writing and not self.is_reading:
            self.close()

        # open
        self._open_read()

        # read
        views = self._get_views(keys, columns)

        return views if factory is None else map(factory, views)

    # -- Iteration

    def iter_arrays(self, arrays_per_chunk=None):
        """Iterates over the arrays in this store."""
        if arrays_per_chunk is None:
            for key in range(self.journal().numarrays()):
                yield self.get([key])
        elif arrays_per_chunk <= 0:
            raise ValueError('arrays_per_chunk must be None or bigger than 0, it is %r' % arrays_per_chunk)
        else:
            for segments in partition_all(arrays_per_chunk, range(self.journal().numarrays())):
                yield self.get(segments)

    def __iter__(self, arrays_per_chunk=None):
        """Alias to iter_arrays."""
        return self.iter_arrays(arrays_per_chunk=arrays_per_chunk)

    # def iter_rows(self, max_rows_per_chunk):
    #     # Iterates segments in chunks with max_rows_per_chunk as upper bound
    #     # (but will give at least one segment at a time)
    #     # This can be more (e.g. SegmentRawStorage) or less involved (e.g. JaggedByNumpy)
    #     # Useful to iterate with really controlled amount of memory
    #     raise NotImplementedError()

    #  --- Factories / curries / partials

    def copyconf(self, **params):
        """Returns a partial function that instantiates this type of store
        with changed default parameters.

        N.B. this default implementation is based on being able to retrieve all default parameters
        using the `what` method; override if that is not the case.

        Parameters
        ----------
        params: **dict
          The parameters that will be fixed in the returned factory function.
        """
        return whatable(partial(self.__class__, **merge(self.what().conf, params)), add_properties=False)

    # --- shape, dtype, order

    @property
    def shape(self):
        """Returns a tuple with the current size of the storage in each dimension."""
        ncols = self.ncols
        return None if ncols is None else (self.nrows, ncols)

    @property
    def dtype(self):
        """Returns the data type of the store."""
        template = self.template()
        return None if template is None else template.dtype

    @property
    def ndims(self):
        """Returns the number of dimensions."""
        # Actually at the moment we only support ndims == 2
        shape = self.shape
        return len(self.shape) if shape is not None else None

    @property
    def ncols(self):
        """Returns the number of columns."""
        template = self.template()
        return None if template is None else template.shape[1]

    @property
    def nrows(self):
        """Returns the number of rows in the store."""
        return self.journal().numrows()

    @property
    def narrays(self):
        """Returns the number or arrays in the store."""
        return self.journal().numarrays()

    @property
    def order(self):
        """Returns 'C' for row major, 'F' for column major."""
        template = self.template()
        if template is None:
            return None
        return 'F' if np.isfortran(template) else 'C'

    # --- Context manager and other magics...

    def __enter__(self):
        return self

    def __exit__(self, *_):
        self.close()

    def __len__(self):
        """Returns the size of the leading dimension."""
        return self.shape[0] if self.shape is not None else 0

    # Also consider register to atexit as a parameter to the constructor


# --- Linear stores (can address arbitrary row segments)


class LinearRawStorage(JaggedRawStore):

    __metaclass__ = ABCMeta  # no harm, lint stops complaining

    def __init__(self, path, journal=None, contiguity=None):
        """
        A linear raw storage can access arbitrary rows using base (row index in the storage) and size
        (number of rows to retrieve).

        Parameters
        ----------
        journal : must quack like JaggedJournal, default None
          see base class

        contiguity : string or None, default None
           indicates the type of contiguity sought for the results; for performance segments retrieval
           does not need to be done in any order
             - 'read': a best effort should be done to leave retrieved segments order-contiguous in memory;
                       this can potentially speed up operations reading these data in the order specified by segments
             - 'write': a best effort should be done to write segments sequentially in memory;
                        this can potentially speed up retrieval
             - 'auto': allow the backend to decide the return flavor;
                       using this the backends can return "lazy" or "cached" arrays
                       (for example, views on memmapped arrays or hdf5 datasets)
             - None: do not force any contiguity nor allow any strange return, just plain numpy arrays
                     owning their own data; this is safest and usually well performing
           usually 'read' can be a good idea for analysis, and 'auto' can have memory saving benefits
           beware that forcing contiguity for speed might lead to memory leaks
           (the whole retrieved segments won't be released while any of them is reacheable)
        """
        super(LinearRawStorage, self).__init__(path, journal=journal)
        self.contiguity = contiguity

    def _get_views(self, keys, columns):
        # get all segments if segments is None
        if keys is None:
            keys = range(self.journal().numarrays())
        keys = [self.journal().base_size(key) if isinstance(key, int) else key for key in keys]

        if 0 == len(keys):
            return []

        # retrieve data
        ne, nc = self.shape
        views = retrieve_contiguous(keys, columns, self._get_hook, self.dtype, ne, nc, self.contiguity)

        return views

    def _get_hook(self, base, size, columns, dest):
        raise NotImplementedError()

    def iter_rows(self, rows_per_chunk):
        """Reads rows_per_chunk rows at a time until all is read."""
        base = 0
        total = len(self)
        while base < total:
            size = min(rows_per_chunk, total - base)
            yield self.get([(base, size)])[0]
            base += size


def retrieve_contiguous(segments, columns, reader, dtype, ne, nc, contiguity):

    # Check for valid contiguity
    if contiguity not in ('read', 'write', 'auto', None):
        raise ValueError('Unknown contiguity scheme: %r' % contiguity)

    # Check query sanity and prepare contiguous query
    # dest_base tells where each query must go to in case of contiguity='read'
    # note that dest_base is not useful for unsorting in the presence of 0-length items (so we explicitly store order)
    dest_base = 0
    query_dest = []
    for order, (base, size) in enumerate(segments):
        if (base + size) > ne or base < 0:
            raise ValueError('Out of bounds query (base=%d, size=%d, maxsize=%d)' % (base, size, ne))
        query_dest.append((order, base, dest_base, size))
        dest_base += size
    total_size = dest_base

    nc = len(columns) if columns is not None else nc

    # Retrieve
    views = []
    if contiguity == 'read':
        # Hope for one-malloc only, but beware of memory leaks
        dest = np.empty((total_size, nc), dtype=dtype)
        # Populate
        for order, base, dest_base, size in sorted(query_dest):
            view = dest[dest_base:dest_base+size]
            view = reader(base, size, columns, view)
            views.append((order, view))
    elif contiguity == 'write':
        # Hope for one-malloc only, but beware of memory leaks
        dest = np.empty((total_size, nc), dtype=dtype)
        # Populate
        dest_base = 0
        for order, base, _, size in sorted(query_dest):
            view = dest[dest_base:dest_base+size]
            view = reader(base, size, columns, view)
            dest_base += size
            views.append((order, view))
    elif contiguity == 'auto':
        for order, base, _, size in sorted(query_dest):
            view = reader(base, size, columns, None)
            views.append((order, view))
    else:
        for order, base, _, size in sorted(query_dest):
            view = np.empty((size, nc), dtype=dtype)
            view = reader(base, size, columns, view)
            views.append((order, view))

    # Unpack views while restoring original order
    return list(map(itemgetter(1), sorted(views, key=itemgetter(0))))



import numpy as np

from whatami import whatable

try:
    import blosc
except ImportError:  # pragma: no cover
    blosc = None


# --- Consistent compressors API


@whatable
class Compressor(object):

    def compress(self, data):
        raise NotImplementedError()

    def decompress(self, cdata):
        raise NotImplementedError()

    def uncompress(self, cdata):  # pragma: no cover
        return self.decompress(cdata)


# --- Pimping Blosc to compress our arrays

class BloscCompressor(Compressor):

    # This has quite an overhead beyond compression ATM

    def __init__(self, shuffle=True, level=5, cname='lz4hc', n_threads=1, dtype=None, shape=None, order=None):
        super(BloscCompressor, self).__init__()
        self.shuffle = shuffle
        self.level = level
        self.cname = cname
        self.n_threads = n_threads
        self._dtype = dtype
        self._shape = shape
        self._order = order

    def compress(self, x):
        blosc.set_nthreads(self.n_threads)  # mmmm global, put in a context to reset
        x = np.ascontiguousarray(x)  # LOOK AT THIS
        shape, dtype, order = x.shape, x.dtype, ('F' if np.isfortran(x) else 'C')
        if self._dtype is None:
            self._shape, self._dtype, self._order = shape, dtype, order
        else:
            assert order == self._order
            assert dtype == self._dtype
            assert len(self._shape) == 1 or shape[1] == self._shape[1]
        return blosc.compress_ptr(x.__array_interface__['data'][0],
                                  x.size, x.dtype.itemsize,
                                  shuffle=self.shuffle, cname=self.cname, clevel=self.level)

    def decompress(self, cx):
        blosc.set_nthreads(self.n_threads)  # mmmm global, put in a context to reset
        x = blosc.decompress(cx)
        x = np.frombuffer(x, dtype=self._dtype)  # beware gets an immutable array
        if self._order == 'F':
            np.asfortranarray(x)  # correct? makes copy and screwes up?
        if len(self._shape) > 1:
            x = x.reshape(-1, self._shape[1])
        return x



# coding=utf-8
from __future__ import print_function, absolute_import, unicode_literals

import numpy as np
import pytest

from functools import partial

from jagged.bcolz_backend import JaggedByCarray
from jagged.blosc_backend import JaggedByBlosc
from jagged.h5py_backend import JaggedByH5Py
from jagged.joblib_backend import JaggedByJoblib
from jagged.mmap_backend import JaggedByMemMap
from jagged.npy_backend import JaggedByNPY
from jagged.bloscpack_backend import JaggedByBloscpack
from jagged.pickle_backend import JaggedByPickle

RAW_STORES = []

LINEAR_RAW_STORES = (
    ('jr=carray', JaggedByCarray),
    ('jr=carraychunks', partial(JaggedByCarray, chunklen=100)),
    ('jr=h5py', JaggedByH5Py),
    ('jr=h5pychunks', partial(JaggedByH5Py, chunklen=100)),
    ('jr=mmap', JaggedByMemMap),
)

for contiguity in ('read', 'write', None, 'auto'):
    for name, store in LINEAR_RAW_STORES:
        RAW_STORES.append((name + '#' + 'cont=%s' % contiguity, partial(store, contiguity=contiguity)))

RAW_STORES.extend([
    ('jr=npy', JaggedByNPY),
    ('jr=blp', JaggedByBloscpack),
    ('jr=blosc-mm', JaggedByBlosc),
    ('jr=pickle', JaggedByPickle),
    ('jr=joblib', JaggedByJoblib),
])

stores = [store for _, store in RAW_STORES]
names = [name for name, _ in RAW_STORES]


@pytest.yield_fixture(params=stores, ids=names)
def jagged_raw(request, tmpdir):
    jr = request.param
    dest = tmpdir.join(jr().what().id()).ensure_dir()
    try:
        yield jr, str(dest)
    finally:
        dest.remove(ignore_errors=True)


@pytest.fixture(params=(1, 2, 10), ids=('ncol=1', 'ncol=2', 'ncol=10'))
def ncol(request):
    return request.param


@pytest.fixture(params=('cols=all', 'cols=all-exp', 'cols=first', 'cols=last',
                        'cols=even', 'cols=inverse', 'cols=mixed'),
                ids=('cols=all', 'cols=all-exp', 'cols=first', 'cols=last',
                     'cols=even', 'cols=inverse', 'cols=mixed'))
def columns(request, ncol):
    if request.param == 'cols=all':
        return None
    elif request.param == 'cols=all-exp':
        return range(ncol)
    elif request.param == 'cols=first':
        return [0]
    elif request.param == 'cols=last':
        return [ncol - 1]  # (we should support python negative indexing syntax)
    elif request.param == 'cols=even':
        return list(range(0, ncol, 2))
    elif request.param == 'cols=inverse':
        return list(range(ncol)[::-1])
    elif request.param == 'cols=mixed':
        list(range(ncol)[::-1]) + list(range(0, ncol, 2))
    else:  # pragma: no cover
        raise ValueError('Unknows column spec %r' % request.param)


@pytest.fixture(params=(0, 1), ids=['rng=0', 'rng=1'])
def rng(request):
    return np.random.RandomState(request.param)


@pytest.fixture
def dataset(ncol, rng):
    sizes = [0] + rng.randint(low=0, high=500, size=10).tolist()
    rng.shuffle(sizes)
    originals = [rng.rand(size, ncol) for size in sizes]
    return rng, originals, ncol


@pytest.fixture
def mock_jagged_raw(dataset):

    # unbox the fixture
    rng, originals, ncol = dataset

    # reader
    jagged = np.vstack(originals)

    def reader(base, size, columns, dest):
        view = jagged[base:(base+size)]
        if columns is not None:
            view = view[:, tuple(columns)]
        if dest is None:
            return view
        dest[:] = view
        return dest

    # shape
    ne, nc = jagged.shape

    # segments
    base = 0
    segments = []
    for o in originals:
        segments.append((base, len(o)))
        base += len(o)

    return originals, ne, nc, originals[0].dtype, segments, reader, rng


@pytest.fixture(params=('read', 'write', None, 'auto'),
                ids=('cont=read', 'cont=write', 'cont=none', 'cont=auto'))
def contiguity(request):
    return request.param



#!/usr/bin/env python
# coding=utf-8

# Authors: Santi Villalba <sdvillal@gmail.com>
# Licence: BSD 3 clause

try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup

import jagged

setup(
    name='jagged',
    license='BSD 3 clause',
    description='Simple tricks for efficient loading or merging collections of unevenly sized elements',
    long_description=open('README.rst').read().replace('|Build Status| |Coverage Status| |Scrutinizer Status|', ''),
    version=jagged.__version__,
    url='https://github.com/sdvillal/jagged',
    author='Santi Villalba',
    author_email='sdvillal@gmail.com',
    packages=['jagged',
              'jagged.compression',
              'jagged.benchmarks',
              'jagged.tests'],
    classifiers=[
        'Intended Audience :: Science/Research',
        'Intended Audience :: Developers',
        'Topic :: Software Development',
        'Topic :: Scientific/Engineering',
        'License :: OSI Approved',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Operating System :: Unix',
    ],
    install_requires=['future',
                      'numpy',
                      'pandas',
                      'whatami>=4.0.0',
                      'toolz',
                      'blosc',
                      'bloscpack',
                      'bcolz',
                      'h5py',
                      'joblib',
                      'psutil'],
    tests_require=['pytest'],
)
